use colored::Colorize;

const KiB : u64 = 1024;
const MiB : u64 = 1024 * KiB;

const L1 : u64 = 32 * KiB;
const L2 : u64 = 512 * KiB;
const L3 : u64 = 16 * MiB;


fn decompose(p: u64) -> (u64, u64) {
    let s = (p as f64).sqrt().log2();
    return (2_u64.pow(s.floor() as u32), 2_u64.pow(s.ceil() as u32));
}

fn report_human(size : u64, n_procs : u64, mem_core : u64, mem_ccx : u64) {
    let (i, j ) = decompose(n_procs);
    let x = size / i;
    let y = size / j;

    println!("{size}^2 / {n_procs} MPI ranks ({x}x{y})");

    let mut L3_stat = "".green();
    let mut L3_usage = (mem_core as f64) / (L3 as f64);

    if mem_core < L3{
        L3_stat = "in L3".green();
    } else if mem_core == L3{
        L3_stat = "maybe in L3".yellow();
    } else {
        L3_stat = "not in L3".red();
    }
    println!("  core mem: {} MiB {} ({})", mem_core/MiB, L3_stat, L3_usage);

    let mut L3_usage = (mem_ccx as f64) / (L3 as f64);
    if mem_ccx < L3{
        L3_stat = "in L3".green();
    } else if mem_ccx == L3{
        L3_stat = "maybe in L3".yellow();
    } else {
        L3_stat = "not in L3".red();
    }

    println!("   ccx mem: {} MiB {} ({})", mem_ccx/MiB, L3_stat, L3_usage);
}


fn report_csv(size : u64, n_procs : u64, mem_core : u64, mem_ccx : u64) {
    let (i, j ) = decompose(n_procs);
    let x = size / i;
    let y = size / j;
    let core_L3_usage = (mem_core as f64) / (L3 as f64);
    let ccx_L3_usage = (mem_ccx as f64) / (L3 as f64);

    println!("{size}, {n_procs}, {x}, {y}, {mem_core}, {core_L3_usage}, {mem_ccx}, {ccx_L3_usage}");
}

fn main() {

    const human : i32 = 0;
    const csv : i32 = 1;

    let mode = human;

    let sizes = [128, 256, 512, 1024, 2048, 4096];
    let proc_counts = [1, 2,4,8,16,32,64,128];

    if mode == csv {
        println!("size,n_procs,x,y,mem_core,core_L3_usage,mem_ccx,ccx_L3_usage");
    }
    for size in sizes {

        if mode == human {
            println!();
        }

        for n_procs in proc_counts {
            
            const mem_datatype : u64 = 8;
            const n_grids : u64 = 4;

            let mem_core = n_grids*mem_datatype*size*size/n_procs;
            let ccx_cores = std::cmp::max(n_procs/32, 1);
            let mem_ccx = mem_core*ccx_cores;
            
            if mode == human {
                report_human(size, n_procs, mem_core, mem_ccx);
            } else {
                report_csv(size, n_procs, mem_core, mem_ccx);
            }
        }
    }

}
